from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy, BaseQuery
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

from . import hard_code
from .config import Config


class Database(SQLAlchemy):
    session: Session

    def dispose_engine(self, **kwargs):
        engine = self.get_engine(**kwargs)  # type: Engine
        engine.dispose()

    @staticmethod
    def query(model):
        query = getattr(model, 'query')  # type: BaseQuery
        return query

    def add_instance(self, instance, commit=True):
        self.session.add(instance)
        if commit:
            self.session.commit()

    def delete_instance(self, instance, commit=True):
        self.session.delete(instance)
        if commit:
            self.session.commit()


@Config.process
def merge_uri(config):
    db = Config.section(hard_code.CK_DATABASE)
    driver = db['driver'].lower()
    if 'mysql' in driver or 'postgresql' in driver:
        uri = '%(driver)s://%(user)s:%(password)s@%(host)s:%(port)d/%(database)s?charset=%(charset)s' % db
    else:
        uri = '%(driver)s://%(path)s' % db

    config['SQLALCHEMY_DATABASE_URI'] = uri


db = Database()
migrate = Migrate()



import re

from flask import render_template

from saika import hard_code
from saika.meta_table import MetaTable
from .web import WebController


class ViewControlller(WebController):
    def assign(self, **kwargs):
        context = self.context.g_get(hard_code.GK_CONTEXT)
        if context is None:
            context = {}
            self.context.g_set(hard_code.GK_CONTEXT, context)
        context.update(kwargs)

    def fetch(self, template=None):
        if template is None:
            view_function = self.context.get_view_function()

            url_prefix = self.options.get('url_prefix').strip('/')  # type: str
            rule_str = MetaTable.get(view_function, hard_code.MK_RULE_STR).strip('/')  # type: str

            if not url_prefix:
                url_prefix = self.name
            if not rule_str:
                rule_str = view_function.__name__

            template = '%s/%s' % (url_prefix, rule_str)
            template = re.sub('<.+?>', '', template)
            template = re.sub('/+', '/', template)
            template = '%s.html' % template.strip('/')

        context = self.context.g_get(hard_code.GK_CONTEXT, {})
        return render_template(template, **context)



import base64

from itsdangerous import TimedJSONWebSignatureSerializer

from .environ import Environ


def obj_encrypt(obj, expires_in=None):
    return TimedJSONWebSignatureSerializer(Environ.app.secret_key, expires_in).dumps(obj).decode()


def obj_decrypt(obj_str):
    try:
        return TimedJSONWebSignatureSerializer(Environ.app.secret_key).loads(obj_str)
    except:
        return None


def obj_standard(obj, str_key=False, str_obj=False):
    this = lambda x: obj_standard(x, str_key, str_obj)
    if type(obj) in [bool, int, float, str, type(None)]:
        return obj
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    elif isinstance(obj, list):
        return [this(i) for i in obj]
    elif isinstance(obj, dict):
        return {str(k) if str_key else this(k): this(v) for k, v in obj.items()}
    else:
        return str(obj) if str_obj else obj



_mt = {}


class MetaTable:
    @staticmethod
    def _item(i):
        if i not in _mt:
            _mt[i] = {}
        return _mt[i]

    @staticmethod
    def set(i, k, v):
        data = MetaTable._item(i)
        data[k] = v

    @staticmethod
    def get(i, k, d=None):
        data = MetaTable._item(i)
        if k not in data and d is not None:
            data[k] = d
        return data.get(k, d)

    @staticmethod
    def all(i):
        return _mt.get(i)

    @staticmethod
    def dump():
        return _mt.copy()



from werkzeug.exceptions import HTTPException

from saika.enums import APP_ERROR


class AppException(HTTPException):
    def __init__(self, error_code=APP_ERROR[0], msg=APP_ERROR[1], data=None, code=500):
        super().__init__(msg, None)
        self.error_code = error_code
        self.msg = msg
        self.data = data
        self.code = code



from flask_wtf import FlaskForm
from werkzeug.datastructures import MultiDict
from wtforms_json import flatten_json

from saika.context import Context


class Form(FlaskForm):
    data: dict
    errors: dict


class ArgsForm(Form):
    def __init__(self, **kwargs):
        super().__init__(MultiDict(Context.request.args), **kwargs)


class JSONForm(Form):
    def __init__(self, **kwargs):
        formdata = Context.request.get_json()
        if formdata is not None:
            formdata = MultiDict(flatten_json(self.__class__, formdata))
        super().__init__(formdata, **kwargs)



from .forms import Form, ArgsForm, JSONForm
from .process import FormException, AUTO


def simple_choices(obj):
    if isinstance(obj, list):
        return [(i, i) for i in obj]
    elif isinstance(obj, dict):
        return [(v, k) for k, v in obj.items()]
    return obj


def set_default_validate(enable=AUTO):
    from saika import hard_code
    from saika.meta_table import MetaTable
    MetaTable.set(hard_code.MI_GLOBAL, hard_code.MK_FORM_VALIDATE, enable)



import json
import os
import sys

from . import hard_code

_config = {}
_processes = []


class Config:
    @staticmethod
    def load(path):
        global _config
        if not os.path.exists(path):
            print('Config not exist: %s' % path, file=sys.stderr)
        else:
            with open(path, 'r') as io:
                _config = json.load(io)

    @staticmethod
    def save(path):
        cfg_str = json.dumps(_config, ensure_ascii=False, indent=2)
        with open(path, 'w') as io:
            io.write(cfg_str)

    @staticmethod
    def section(key):
        cfg = _config.get(key, {})  # type: dict
        return cfg

    @staticmethod
    def all():
        return _config

    @staticmethod
    def merge():
        config = {}

        for process in _processes:
            process(config)

        config.update(_config.get(hard_code.CK_CORE))
        return config

    @staticmethod
    def process(f):
        _processes.append(f)
        return f



from saika import hard_code
from saika.meta_table import MetaTable


def rule(rule_str: str):
    def wrapper(f):
        MetaTable.set(f, hard_code.MK_RULE_STR, rule_str)
        return f

    return wrapper


def rule_rs(rule_str: str):
    def wrapper(f):
        MetaTable.set(f, hard_code.MK_RULE_STR, rule_str.rstrip('/'))
        return f

    return wrapper


def controller(url_prefix, template_folder=None, static_folder=None, **options):
    opts = locals().copy()
    opts.update(opts.pop('options'))

    def wrapper(cls):
        controllers = MetaTable.get(hard_code.MI_GLOBAL, hard_code.MK_CONTROLLER_CLASSES, [])  # type: list
        controllers.append(cls)
        MetaTable.set(cls, hard_code.MK_OPTIONS, opts)
        return cls

    return wrapper


def _method(f, method):
    methods = MetaTable.get(f, hard_code.MK_METHODS, [])  # type: list
    methods.append(method)
    return f


get = lambda f: _method(f, 'GET')
post = lambda f: _method(f, 'POST')



from saika import hard_code
from saika.meta_table import MetaTable


def _callback(f, key):
    callbacks = MetaTable.get(hard_code.MI_CALLBACK, key, [])  # type: list
    callbacks.append(f)
    return f


before_app_request = lambda f: _callback(f, hard_code.MK_BEFORE_APP_REQUEST)
before_app_first_request = lambda f: _callback(f, hard_code.MK_BEFORE_APP_FIRST_REQUEST)
after_app_request = lambda f: _callback(f, hard_code.MK_AFTER_APP_REQUEST)



import json

from saika import common
from .exception import AppException


class APIException(AppException):
    def __init__(self, *args, **kwargs):
        kwargs['code'] = 200
        super().__init__(*args, **kwargs)

    def get_body(self, environ=None):
        return json.dumps(
            common.obj_standard(dict(
                code=self.error_code,
                msg=self.msg,
                data=self.data,
            ), True, True))



from flask import abort, redirect, flash, url_for, send_file, send_from_directory, make_response, Flask

from saika import hard_code
from saika.context import Context
from saika.form import Form
from .base import ControllerBase


class WebController(ControllerBase):
    def __init__(self):
        super().__init__()

        self.abort = abort
        self.redirect = redirect
        self.flash = flash
        self.url_for = url_for
        self.send_file = send_file
        self.send_from_directory = send_from_directory
        self.make_response = make_response

    @property
    def form(self):
        form = Context.g_get(hard_code.GK_FORM)  # type: Form
        return form

    def instance_register(self, app: Flask):
        self.callback_before_register()
        app.register_blueprint(self.blueprint, **self.options)



import builtins
import importlib
import os
import pkgutil
import re
import sys
import traceback

from flask import Flask

from . import hard_code
from .config import Config
from .const import Const
from .context import Context
from .controller import WebController
from .cors import cors
from .database import db, migrate
from .environ import Environ
from .meta_table import MetaTable
from .socket import socket, SocketController
from .socket_io import socket_io, SocketIOController


class SaikaApp(Flask):
    def __init__(self, **kwargs):
        super().__init__(self.__class__.__module__, **kwargs)

        self.web_controllers = []
        self.socket_controllers = []
        self.sio_controllers = []

        try:
            self._init_env()
            self._init_config()
            self._init_app()

            self._import_modules()
            self._init_callbacks()
            self._init_context()
            self._init_controllers()
        except:
            traceback.print_exc(file=sys.stderr)

    def _init_env(self):
        if Environ.app is not None:
            raise Exception('SaikaApp was created.')

        Environ.app = self
        Environ.debug = os.getenv(hard_code.SAIKA_DEBUG)
        Environ.program_path = os.path.join(self.root_path, '..')
        Environ.config_path = os.path.join(Environ.program_path, Const.config_file)
        Environ.data_path = os.path.join(Environ.program_path, Const.data_dir)

    def _init_config(self):
        Config.load(Environ.config_path)
        cfg = Config.merge()
        self.config.from_mapping(cfg)

    def _init_app(self):
        db.init_app(self)
        migrate.init_app(self, db)
        cors.init_app(self, **(Config.section('cors') or dict(
            supports_credentials=True,
        )))
        socket_io.init_app(self, **(Config.section('socket_io') or dict(
            cors_allowed_origins='*',
        )))
        socket.init_app(self)
        self.callback_init_app()

    def _init_callbacks(self):
        for f in MetaTable.get(hard_code.MI_CALLBACK, hard_code.MK_BEFORE_APP_REQUEST, []):
            self.before_request(f)
        for f in MetaTable.get(hard_code.MI_CALLBACK, hard_code.MK_BEFORE_APP_FIRST_REQUEST, []):
            self.before_first_request(f)
        for f in MetaTable.get(hard_code.MI_CALLBACK, hard_code.MK_AFTER_APP_REQUEST, []):
            self.after_request(f)

    def _init_controllers(self):
        controller_classes = MetaTable.get(hard_code.MI_GLOBAL, hard_code.MK_CONTROLLER_CLASSES, [])
        for cls in controller_classes:
            if issubclass(cls, WebController):
                item = cls()
                item.instance_register(self)
                self.web_controllers.append(item)
            elif issubclass(cls, SocketController):
                item = cls()
                item.instance_register(socket)
                self.socket_controllers.append(item)
            elif issubclass(cls, SocketIOController):
                options = MetaTable.get(cls, hard_code.MK_OPTIONS)
                item = cls(namespace=options.pop('url_prefix', None))
                socket_io.on_namespace(item)
                self.sio_controllers.append(item)

    def _init_context(self):
        for name, obj in self.make_context().items():
            self.add_template_global(obj, name)

        items = []
        for key in dir(builtins):
            item = getattr(builtins, key)
            type_name = type(item).__name__
            if key[0] != '_' and hasattr(item, '__name__') and (
                    type_name == 'builtin_function_or_method' or re.match('^[a-z]+$', key)):
                items.append(item)

        for item in items:
            self.add_template_global(item)

    def _import_modules(self):
        module = self.__class__.__module__
        sub_modules = list(pkgutil.iter_modules([module], '%s.' % module))
        sub_modules = [i.name for i in sub_modules if i.ispkg]
        for i in sub_modules:
            importlib.import_module(i)

    def callback_init_app(self):
        pass

    @staticmethod
    def make_context():
        context = dict(Config=Config, Const=Const, Context=Context, db=db, Environ=Environ, MetaTable=MetaTable)
        classes = MetaTable.get(hard_code.MI_GLOBAL, hard_code.MK_MODEL_CLASSES, [])
        for cls in classes:
            context[cls.__name__] = cls
        return context



import re

from flask_socketio import Namespace

from saika.context import Context


class SocketIOController(Namespace):
    def trigger_event(self, event, *args):
        event = re.sub('[A-Z]+', lambda x: '_%s' % x.group().lower(), event)
        event = re.sub('[^a-z0-9$]', '_', event)
        event = event.strip('_')
        super().trigger_event(event, *args)

    @property
    def context(self):
        return Context

    @property
    def sid(self):
        return Context.request.sid

    def on_connect(self):
        pass

    def on_disconnect(self):
        pass

    def emit(self, event, data=None, broadcast=False, room=None, **kwargs):
        if broadcast:
            room = None
        elif room is None:
            room = self.sid

        return super().emit(event, data, room, **kwargs)

    def send(self, data=None, broadcast=False, room=None, **kwargs):
        if broadcast:
            room = None
        elif room is None:
            room = self.sid

        return super().send(data, room, **kwargs)

    def disconnect(self, **kwargs):
        self._inject_kwargs(kwargs)
        return super().disconnect(**kwargs)

    def rooms(self, **kwargs):
        self._inject_kwargs(kwargs)
        return super().rooms(**kwargs)

    def enter_room(self, room, **kwargs):
        self._inject_kwargs(kwargs)
        return super().enter_room(room=room, **kwargs)

    def leave_room(self, room, **kwargs):
        self._inject_kwargs(kwargs)
        return super().leave_room(room=room, **kwargs)

    def _inject_kwargs(self, kwargs: dict):
        kwargs['sid'] = kwargs.get('sid')
        if not kwargs['sid']:
            kwargs['sid'] = self.sid
        kwargs['namespace'] = kwargs.get('namespace', self.namespace)



from flask import Flask


class Environ:
    app: Flask = None
    debug: bool
    program_path: str
    config_path: str
    data_path: str

    @staticmethod
    def into_app_context_do(f, *args, **kwargs):
        with Environ.app.app_context():
            return f(*args, **kwargs)

    @staticmethod
    def into_request_context_do(environ, f, *args, **kwargs):
        with Environ.app.request_context(environ):
            return f(*args, **kwargs)



import json

from geventwebsocket import WebSocketError
from geventwebsocket.websocket import WebSocket, MSG_SOCKET_DEAD

from saika import hard_code, common
from .controller import SocketController


class EventSocketController(SocketController):
    @property
    def _loop(self):
        socket = self.socket
        if isinstance(socket, WebSocket):
            return not socket.closed

    def handle(self, socket: WebSocket):
        self.context.g_set(hard_code.GK_SOCKET, socket)

        self.on_connect()
        while self._loop:
            data_str = None
            try:
                data_str = socket.receive()
                if not data_str:
                    continue
                data = json.loads(data_str)  # type: dict
                if isinstance(data, dict):
                    event = 'on_%s' % data.pop('event')
                    if hasattr(self, event) and event not in dir(EventSocketController):
                        kwargs = data.pop('data', {})
                        getattr(self, event)(**kwargs)
                        continue
                self.on_message(data)
            except json.JSONDecodeError:
                self.on_message(data_str)
            except WebSocketError as e:
                if MSG_SOCKET_DEAD in e.args:
                    break
                self.on_error(e)
            except Exception as e:
                self.on_error(e)
        self.on_disconnect()

    @property
    def socket(self):
        socket = self.context.g_get(hard_code.GK_SOCKET)  # type: WebSocket
        return socket

    def send(self, data: dict):
        self.socket.send(json.dumps(common.obj_standard(data, True, True)))

    def emit(self, event: str, data: dict):
        self.send(dict(event=event, data=data))

    def disconnect(self, *args, **kwargs):
        self.socket.close(*args, **kwargs)

    def on_connect(self):
        pass

    def on_disconnect(self):
        pass

    def on_message(self, data):
        pass

    def on_error(self, e: Exception):
        raise e



from . import hard_code
from .workers import worker, gevent_patch

if not __import__('os').getenv(hard_code.SAIKA_DEBUG):
    gevent_patch()

from .app import SaikaApp
from .config import Config
from .const import Const
from .context import Context
from .controller import WebController, APIController, ViewControlller
from .cors import cors
from .database import db, migrate
from .environ import Environ
from .exception import AppException, APIException
from .manager import init_manager
from .meta_table import MetaTable
from .socket import socket, SocketController, EventSocketController
from .socket_io import socket_io, SocketIOController



SAIKA_DEBUG = 'SAIKA_DEBUG'

# region Meta Item
MI_GLOBAL = 'global'
MI_CALLBACK = 'callback'
# endregion

# region Meta Key

# region Global Part
MK_CONTROLLER_CLASSES = 'controller_classes'
MK_MODEL_CLASSES = 'model_classes'
# endregion

# region Function Key
MK_RULE_STR = 'rule_str'
MK_METHODS = 'methods'
MK_OPTIONS = 'options'
MK_FORM_CLASS = 'form_class'
MK_FORM_ARGS = 'form_args'
MK_FORM_VALIDATE = 'form_validate'
# endregion

# region Config Key
CK_CORE = 'core'
CK_DATABASE = 'database'
CK_CORS = 'cors'
CK_SOCKET_IO = 'socket_io'
# endregion

# region Arg Key
AK_VALIDATE = 'validate'
# endregion

# region G Key
GK_KEYS = '_keys'
GK_CONTEXT = 'context'
GK_FORM = 'form'
GK_SOCKET = 'socket'
# endregion

# region Callback Key
MK_BEFORE_APP_REQUEST = 'before_app_request'
MK_BEFORE_APP_FIRST_REQUEST = 'before_app_first_request'
MK_AFTER_APP_REQUEST = 'after_app_request'
# endregion

# endregion



import sys

from flask_migrate import MigrateCommand
from flask_script import Manager, Server
from werkzeug.serving import is_running_from_reloader

from .app import SaikaApp
from .const import Const
from .environ import Environ
from .socket_io import socket_io


class GEventServer(Server):
    def __call__(self, app: SaikaApp, host, port, use_debugger, use_reloader,
                 threaded, processes, passthrough_errors, ssl_crt, ssl_key):
        kwargs = dict(debug=use_debugger, use_reloader=use_reloader, certfile=ssl_crt, keyfile=ssl_key)
        for k, v in list(kwargs.items()):
            if v is None:
                kwargs.pop(k)

        if use_reloader is None:
            use_reloader = app.debug

        if not use_reloader or is_running_from_reloader():
            print(' * Serving Saika "%s"' % (Environ.app.import_name))
            print('   - Saika Version: %s' % Const.version)
            print(' * Environment: %s' % app.env)
            print(' * Debug mode: %s' % ('on' if app.debug else 'off'))
            print(' * Running on http://%s:%s/ (Press CTRL+C to quit)' % (host, port))

        socket_io.server.eio.async_mode = 'gevent'
        socket_io.run(app, host, port, log_output=True, **kwargs)


def init_manager(app: SaikaApp, **kwargs):
    manager = Manager(app, **kwargs)
    manager.add_command('db', MigrateCommand)
    manager.shell(app.make_context)
    if not Environ.debug:
        manager.add_command('runserver', GEventServer())
    else:
        print(' * Saika Debug: Websocket is disabled now.', file=sys.stderr)

    return manager



from saika.controller import ControllerBase


class SocketController(ControllerBase):
    def instance_register(self, socket):
        self.callback_before_register()
        socket.register_blueprint(self.blueprint, **self.options)



import re

from flask import Blueprint

from saika import hard_code
from saika.context import Context
from saika.environ import Environ
from saika.meta_table import MetaTable


class ControllerBase:
    def __init__(self):
        name = self.__class__.__name__.replace('Controller', '')
        self._name = re.sub('[A-Z]', lambda x: '_' + x.group().lower(), name).lstrip('_')
        self._import_name = self.__class__.__module__

        self._blueprint = Blueprint(self._name, self._import_name)
        self._register_methods()

    @property
    def blueprint(self):
        return self._blueprint

    @property
    def name(self):
        return self._name

    @property
    def context(self):
        return Context

    @property
    def request(self):
        return Context.request

    @property
    def options(self):
        options = MetaTable.get(self.__class__, hard_code.MK_OPTIONS, {})  # type: dict
        return options

    @property
    def view_function_options(self):
        options = MetaTable.all(Context.get_view_function())  # type: dict
        return options

    def _register_methods(self):
        if Environ.debug:
            print(' * Init %s (%s): %a' % (self._import_name, self._name, self.options))

        keeps = dir(ControllerBase)
        for k in dir(self):
            if k in keeps:
                continue

            t = getattr(self.__class__, k, None)
            if isinstance(t, property):
                continue

            _f = f = getattr(self, k)
            if callable(f):
                if hasattr(f, '__func__'):
                    f = f.__func__
                meta = MetaTable.all(f)
                if meta is not None:
                    options = dict()
                    methods = meta.get(hard_code.MK_METHODS)
                    if methods:
                        options['methods'] = methods

                    self._blueprint.add_url_rule(meta[hard_code.MK_RULE_STR], None, _f, **options)
                    if Environ.debug:
                        name = _f.__name__
                        if hasattr(_f, '__qualname__'):
                            name = _f.__qualname__
                        print('   - %s: %a' % (name, options))

    def instance_register(self, *args, **kwargs):
        pass

    def callback_before_register(self):
        pass



#!/usr/bin/env python
import setuptools

from saika.const import Const

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.read().splitlines()

setuptools.setup(
    name="Saika",
    version=Const.version,
    author="HsOjo",
    author_email="hsojo@qq.com",
    keywords='hsojo python3 flask web',
    description='''A simple web framework base on Flask.''',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HsOjo/Saika/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License (GPL)",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
)



from saika import hard_code
from saika.meta_table import MetaTable


def model(cls):
    models = MetaTable.get(hard_code.MI_GLOBAL, hard_code.MK_MODEL_CLASSES, [])  # type: list
    models.append(cls)
    return cls



from flask import request, session, g, current_app

from saika import hard_code


class Context:
    request = request
    session = session
    current_app = current_app
    g = g

    @staticmethod
    def g_set(k, v):
        keys = Context.g_get(hard_code.GK_KEYS)
        if keys is None:
            keys = []
            setattr(g, hard_code.GK_KEYS, keys)

        keys.append(k)
        setattr(g, k, v)

    @staticmethod
    def g_get(k, default=None):
        return getattr(g, k, default)

    @staticmethod
    def g_all():
        r = {}
        for k in Context.g_get(hard_code.GK_KEYS, []):
            r[k] = Context.g_get(k)
        return r

    @staticmethod
    def get_view_function():
        f = current_app.view_functions.get(request.endpoint)
        if hasattr(f, '__func__'):
            f = f.__func__

        return f



from saika import hard_code
from saika.meta_table import MetaTable


def form(form_cls, validate=None, **kwargs):
    if validate is None:
        validate = MetaTable.get(hard_code.MI_GLOBAL, hard_code.MK_FORM_VALIDATE)

    kwargs[hard_code.AK_VALIDATE] = validate

    def wrapper(f):
        MetaTable.set(f, hard_code.MK_FORM_CLASS, form_cls)
        MetaTable.set(f, hard_code.MK_FORM_ARGS, kwargs)
        return f

    return wrapper



import sys
import traceback

from flask import jsonify

from saika.exception import AppException, APIException
from .web import WebController


class APIController(WebController):
    def callback_before_register(self):
        @self.blueprint.errorhandler(AppException)
        def convert(e: AppException):
            return APIException(e.error_code, e.msg, e.data)

        @self.blueprint.errorhandler(Exception)
        def catch(e: Exception):
            traceback.print_exc(file=sys.stderr)
            return APIException(data=dict(exc=str(e)))

    def response(self, code=0, msg=None, **data):
        return jsonify(code=code, msg=msg, data=data)

    def success(self, code=0, msg=None, **data):
        raise APIException(code, msg, data)

    def error(self, code=1, msg=None, **data):
        raise APIException(code, msg, data)



import os
import sys

worker = None

try:
    from geventwebsocket.gunicorn.workers import GeventWebSocketWorker
    from gunicorn.workers.base_async import AsyncWorker


    class SaikaWorker(GeventWebSocketWorker):
        def notify(self):
            super(AsyncWorker, self).notify()
            # Fix fork child process from self.
            ppid = os.getppid()
            if ppid not in [self.ppid, self.pid]:
                self.log.info("Parent changed, shutting down: %s", self)
                sys.exit(0)


    worker = SaikaWorker
except ImportError:
    pass


def gevent_patch():
    try:
        from gevent import monkey

        monkey.patch_all()
    except ImportError:
        pass



from wtforms_json import init

from saika import hard_code, common
from saika.context import Context
from saika.decorator.request import before_app_request
from saika.enums import PARAMS_MISMATCH
from saika.exception import AppException
from saika.meta_table import MetaTable

init()
AUTO = object()


class FormException(AppException):
    pass


@before_app_request
def process_form():
    if Context.request.method == 'OPTIONS':
        return

    f = Context.get_view_function()
    cls = MetaTable.get(f, hard_code.MK_FORM_CLASS)
    if cls is not None:
        kwargs = MetaTable.get(f, hard_code.MK_FORM_ARGS).copy()  # type: dict

        validate = kwargs.pop(hard_code.AK_VALIDATE)
        validate = AUTO if validate is None else validate

        form = cls(**kwargs)
        Context.g_set(hard_code.GK_FORM, form)

        if validate is AUTO:
            validate = Context.request.method != 'GET'
        if validate and not form.validate():
            raise FormException(*PARAMS_MISMATCH, data=dict(
                errors=common.obj_standard(form.errors, True, True)
            ))
